from appdirs import AppDirs

from os.path import join, exists
from os import makedirs


DIRS = AppDirs('lanku', 'LANNOCC')

if not exists(DIRS.user_data_dir):
    makedirs(DIRS.user_data_dir)

def _load_(filename):
    config = join(DIRS.user_data_dir, filename)
    if not exists(config):
        return None

    with open(config, 'r') as config:
        return config.read()

def _save_(filename, txt):
    config = join(DIRS.user_data_dir, filename)

    with open(config, 'w') as config:
        config.write(txt)

def load_win_config(name):
    config = _load_(name)
    if config is None: return None
    return [ int(x) for x in config.strip().split(',') ]

def save_win_config(name, x, y, w, h):
    _save_(name, f'{x},{y},{w},{h}')

def load_connect_label():
    return _load_('label')

def save_connect_label(label):
    _save_('label', label)

def load_interests():
    config = _load_('interests')
    if config is None: return [ ]
    return config.split('\n')

def save_interests(interests):
    _save_('interests', '\n'.join(interests))




from lank.peer import Master as Base
from lank.crypto import get_handler as get_crypto
from requests import get
#from miniupnpc import UPnP


class Master(Base):
    def __init__(self, app):
        super().__init__(int(app.home.pnl_config.peer_port.value))
        self.app = app

    def get_public_ip(self):
        return get('https://api.ipify.org').content.decode('utf-8')

    '''
    def upnp(self):
        upnp = UPnP()
        upnp.discoverdelay = 10
        upnp.discover()
        upnp.selectigd()
        upnp.addportmapping(self.port, 'TCP', upnp.lanaddr,
                            self.port, 'lanku', '')
    '''




from . import config

from lank.node import NODES, HELLO_TIMEOUT, GENERAL_TIMEOUT, KEEPALIVE
from lank.node.protocol.v2 import *
from lank.crypto import get_handler as get_crypto
from requests import get

from threading import Thread
from queue import Queue, Empty
import socket
import sys


class Client(Thread):
    def __init__(self, app, verbose=True):
        super().__init__(name='node-client')
        self.app = app
        self.verbose = verbose

        self.label = self.app.home.pnl_config.label.value
        self.pwd = self.app.home.pnl_config.pwd.value
        self.host = self.get_public_ip()
        self.port = int(self.app.home.pnl_config.peer_port.value)

        self.node = None
        self.ping = None
        self.input = Queue(maxsize=1)
        self.output = Queue()

        self.sender_thread = Thread(
            name='node-client-sender', target=self.sender)
        self.receiver_thread = Thread(
            name='node-client-receiver', target=self.receiver)

    def run(self):
        self.print(' - connecting to node:')

        for addr in NODES:
            self.print(f'   * trying {addr}... ', False)

            try:
                sock = socket.create_connection(addr, timeout=HELLO_TIMEOUT)
                sock.settimeout(HELLO_TIMEOUT)
                self.node = Handler(sock, addr)
                self.node.hello()
                self.node.sock.settimeout(GENERAL_TIMEOUT)
                self.print('[READY]')
                break

            except ConnectionRefusedError:
                self.print('[REFUSED]')
                self.node = None

            except socket.timeout:
                self.print('[TIMEOUT]')
                self.node = None

        if not self.node: return
        self.receiver_thread.start()
        self.sender_thread.start()

        self.input.put_nowait(GetRegistration(self.label))
        msg = self.output.get(timeout=GENERAL_TIMEOUT)
        if not msg: return
        if not isinstance(msg, Registration):
            return self.error(f'failed to get label info: {msg}')

        try:
            crypto = get_crypto(msg.version)
            priv_key = crypto.load_private_key(msg.key_pair_pem, self.pwd)

        except TypeError as e:
            return self.error(f'failed to unlock private key: {e}')
        except ValueError as e:
            return self.error(f'failed to unlock private key: {e}')

        msg = PeerOn(crypto.VERSION, self.label, self.host, self.port)
        msg.signature = crypto.sign(priv_key, msg.to_sign(crypto))

        self.input.put_nowait(msg)
        msg = self.output.get(timeout=GENERAL_TIMEOUT)
        if not msg: return
        if not isinstance(msg, Signed):
            return self.error(f'failed to sign on: {msg}')

        config.save_connect_label(self.label)

        self.input.put_nowait(ListLabels())
        msg = self.output.get(timeout=GENERAL_TIMEOUT)
        if not msg: return
        if not isinstance(msg, LabelsList):
            return self.error(f'failed to get labels list: {msg}')

        self.print(f'labels: {msg.labels}')
        del msg.labels[msg.labels.index(self.label)]
        self.app.home.pnl_labels.refresh_labels(msg.labels)

        for label in msg.labels:
            if label in self.app.home.pnl_labels.interests:
                self.input.put(LabelInterest(label))

    def notify(self, label, notify=True):
        self.input.put_nowait(
            LabelInterest(label) if notify else LabelIgnore(label))

    def stop(self):
        if not self.node: return
        node = self.node
        self.node = None

        node.sock.shutdown(socket.SHUT_RDWR)
        node.sock.close()
        self.output.put_nowait(None)
        self.input.put_nowait(None)

        self.app.home.pnl_labels.refresh_labels()

    def join(self):
        self.receiver_thread.join()
        super().join()

    def sender(self):
        try:
            while self.node:
                try:
                    #while msg := self.input.get(timeout=KEEPALIVE):
                    msg = self.input.get(timeout=KEEPALIVE)
                    if msg:
                        if not self.node: break
                        self.send(msg)

                except Empty:
                    if self.ping:
                        return self.error('ping sent but no pong received')

                    if self.node:
                        self.ping = Ping()
                        self.send(self.ping)

        finally:
            self.stop()

    def send(self, msg):
        if not self.node: return
        self.print(f'N    {self.node.addr} <- {msg}')
        self.node.send(msg)

    def receiver(self):
        try:
            #while msg := self.recv():
            msg = self.recv()
            while msg:
                if isinstance(msg, Pong):
                    if not self.ping:
                        return self.error(f'received pong without ping: {msg}')

                    if msg.nonce != self.ping.nonce:
                        return self.error(f'ping-pong nonce mismatch')

                    self.ping = None

                elif isinstance(msg, Signed) and msg.label != self.label:
                    self.app.notify(msg)

                else:
                    self.output.put(msg)

                msg = self.recv()

            self.error('lost connection')

        except OSError as e:
            self.error(e)

        finally:
            self.stop()

    def recv(self):
        if not self.node: return None
        msg = self.node.recv()
        if msg: self.print(f'N    {self.node.addr} -> {msg}')
        return msg

    def get_public_ip(self):
        ip = get('https://api.ipify.org').content.decode('utf-8')
        self.print(f' - our public IP address: {ip}')
        return ip

    def print(self, msg, newline=True):
        if not self.verbose: return

        if newline:
            print(msg)

        else:
            print(msg, end='')
            sys.stdout.flush()

    def error(self, e):
        if not self.node: return
        print(f'** ERROR ** {e}')
        self.stop()




from .config import *

from pyglet.window import Window as PWin
from pyglet.graphics import Batch
from pyglet.shapes import Rectangle
from pyglet.image import ImageData
from pyglet.text import Label as PLabel
from pyglet.gui import Frame as PFrame, PushButton, TextEntry as PTextEntry
from pyglet import clock
import pyglet.resource
pyglet.resource.path = ['@lanku']
pyglet.resource.reindex()

import platform


def image(name):
    return pyglet.resource.image(f'assets/{name}')


class LWin(PWin):
    def __init__(self, app, caption, w, h, save_name=None, resizable=True):
        self.app = app
        self.save_name = save_name

        super().__init__(w, h, caption=caption, resizable=resizable,
            visible=False)

        self.xywh = [0, 0, w, h]
        self.load()
        self.hidden = False
        self.minimized = False

        icon = self.app.icon
        self.set_icon(ImageData(icon.width, icon.height, 'RGBA',
            icon.tobytes(), pitch=-icon.width*4))

        self.frame = PFrame(self)
        self.batch = Batch()
        self.components = [ ]

        self.offset_for = None
        self.offset = None

    def _add_(self, child):
        self.components.append(child)

    def on_draw(self):
        self.clear()
        self.batch.draw()
        for child in self.components:
            child.draw()

    def save(self):
        if not self.save_name or not self.xywh: return
        save_win_config(self.save_name,
                        self.xywh[0], self.xywh[1],
                        self.xywh[2], self.xywh[3])

    def load(self):
        if not self.save_name: return
        config = load_win_config(self.save_name)
        if not config: return

        #print(f'load: {config}')
        self.xywh = config

    def place(self):
        if not self.xywh: return
        #print(f'location set: {self.xywh[0]}, {self.xywh[1]}')
        x = self.xywh[0]
        y = self.xywh[1]

        if self.offset:
            x += self.offset[0]
            y += self.offset[1]
        else:
            self.offset_for = (x, y)

        #print(f'actual set_location: {x}, {y}')
        self.set_size(self.xywh[2], self.xywh[3])
        self.set_location(x, y)

        if not self.offset:
            clock.schedule_once(self.place_offset, 0.5)

    def place_offset(self, dt):
        #print(f'location get: {self.get_location()}')
        x = self.offset_for[0]
        y = self.offset_for[1]

        #if x != self.offset_for[0] or y != self.offset_for[1]:
        #    print('bail')
        #    return

        pos = self.get_location()
        self.offset = (x - pos[0], y - pos[1])
        #print(f'computed window offset: {self.offset}')

        if any(self.offset):
            #self.place()
            self.xywh[0] += self.offset[0]
            self.xywh[1] += self.offset[1]
            #print(f'adjusted xywh: {self.xywh}')

        self.offset_for = None

    def null(self, dt):
        #print('null')
        pass

    def set_visible(self, visible=True):
        #print(f'set_visible {visible}')
        self.hidden = not visible
        restore = self.minimized and visible

        if not visible:
            self.save()

        super().set_visible(visible)

        if restore:
            #print('restore')

            if platform.system() == 'Windows':
                # hack: only way to restore on Windows?
                xywh = list(self.xywh)
                self.maximize()
                self.xywh = xywh

        if visible:
            self.place()

            if not restore:
                # hack: sometimes necessary to trigger redraw on Linux
                clock.schedule_once(self.null, 0.1)

    def on_show(self):
        #print('show')
        self.minimized = False

    def on_hide(self):
        #print('hide')
        if not self.hidden:
            #print('minimized')
            self.minimized = True

    def on_move(self, x, y):
        #print(f'move {x},{y}')
        if x < 0 or y < 0: return

        if not self.offset_for and self.offset:
            x += self.offset[0]
            y += self.offset[1]

        if self.xywh:
            self.xywh[0] = x
            self.xywh[1] = y
        else:
            w, h = self.get_size()
            self.xywh = [x, y, w, h]

    def on_resize(self, w, h):
        #print(f'resize {w},{h}')
        if self.xywh:
            self.xywh[2] = w
            self.xywh[3] = h
        else:
            x, y = self.get_location()
            self.xywh = [x, y, w, h]
        super().on_resize(w, h)

    def on_close(self):
        self.save()
        super().on_close()


class LFrame(PFrame):
    def __init__(self, win, enabled=True, cell_size=64, order=0):
        self.enabled = enabled
        super().__init__(win, cell_size, order)
        self.stack = [ ]

    def _add_(self, child):
        self.stack.append(child)

    def _del_(self, child):
        del self.stack[self.stack.index(child)]

    def enable(self):
        self.enabled = True
        for child in self.stack:
            if child.visible:
                child.cframe.enable()

    def disable(self):
        self.enabled = False
        for child in self.stack:
            if child.visible:
                child.cframe.disable()

    def on_mouse_press(self, x, y, btns, mods):
        if not self.enabled: return
        super().on_mouse_press(x, y, btns, mods)

    def on_mouse_release(self, x, y, btns, mods):
        if not self.enabled: return
        super().on_mouse_release(x, y, btns, mods)

    def on_mouse_drag(self, x, y, dx, dy, btns, mods):
        if not self.enabled: return
        super().on_mouse_drag(x, y, dx, dy, btns, mods)

    def on_mouse_scroll(self, x, y, index, direction):
        if not self.enabled: return
        super().on_mouse_scroll(x, y, index, direction)

    def on_mouse_motion(self, x, y, dx, dy):
        if not self.enabled: return
        super().on_mouse_motion(x, y, dx, dy)

    def on_text(self, text):
        if not self.enabled: return
        super().on_text(text)

    def on_text_motion(self, motion):
        if not self.enabled: return
        super().on_text_motion(motion)

    def on_text_motion_select(self, motion):
        if not self.enabled: return
        super().on_text_motion_select(motion)


class Component:
    def __init__(self, parent, x, y, visible=True):
        self.parent = parent
        if isinstance(self.parent, LWin):
            self.win = self.parent
        else:
            self.win = self.parent.win
        self.pos = [x, y] # relative
        self._visible_ = visible

        self.cbatch = Batch()
        self.cframe = LFrame(self.win, self.visible)

        self.stack = [ ]
        self.parent._add_(self)

    def _add_(self, child):
        self.stack.append(child)
        self.cframe._add_(child)

    def _del_(self, child):
        del self.stack[self.stack.index(child)]
        self.cframe._del_(child)

    @property
    def visible(self):
        return self._visible_

    @visible.setter
    def visible(self, visible):
        self._visible_ = visible
        if visible:
            self.cframe.enable()
        else:
            self.cframe.disable()

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def draw(self):
        if not self.visible: return
        self.cbatch.draw()
        for child in self.stack:
            child.draw()

    def get_abs_pos(self):
        if isinstance(self.parent, LWin):
            return self.pos

        else:
            pos = self.parent.get_abs_pos()
            return [pos[0] + self.pos[0], pos[1] + self.pos[1]]

    def null(self, dt):
        pass


class Label(Component, PLabel):
    def __init__(self, parent, x, y, text):
        Component.__init__(self, parent, x, y)
        pos = self.get_abs_pos()
        PLabel.__init__(self, text, x=pos[0], y=pos[1], batch=self.cbatch)
        #self.cframe.add_widget(self)


class TextEntry(Component, PTextEntry):
    def __init__(self, parent, x, y, w, text=None, mask=None):
        Component.__init__(self, parent, x, y)
        pos = self.get_abs_pos()
        PTextEntry.__init__(self, text, pos[0], pos[1], w,
            caret_color=(0, 255, 0), batch=self.cbatch)
        self.mask = mask

        if self.mask:
            self.real_val = ''
            self._del_text_orig_ = self._doc.delete_text
            self._doc.delete_text = self._del_text_

        self.cframe.add_widget(self)

    def on_mouse_motion(self, x, y, dx, dy):
        if not self.enabled:
            return

        self._set_focus(self._check_hit(x, y))

    @property
    def value(self):
        if self.mask:
            return self.real_val

        else:
            return super().value

    def on_text(self, text):
        if not self.enabled: return

        if self.mask:
            pos = self._caret.position
            self.real_val = self.real_val[:pos] + text + self.real_val[pos:]
            text = self.mask

        super().on_text(text)

    def _del_text_(self, start, end):
        self.real_val = self.real_val[:start] + self.real_val[end:]
        self._del_text_orig_(start, end)


class Button(Component, PushButton):
    def __init__(self, parent, x, y, normal, pushed, hovering):
        Component.__init__(self, parent, x, y)
        pos = self.get_abs_pos()
        PushButton.__init__(self, pos[0], pos[1], pushed, normal, hovering,
            batch=self.cbatch)
        self.cframe.add_widget(self)

    def set_pos(self, x, y):
        self.pos = [x, y]
        pos = self.get_abs_pos()
        self._x = pos[0]
        self._y = pos[1]
        self._update_position()

    def remove(self):
        self.cframe.remove_widget(self)
        self.parent._del_(self)


class ToggleButton(Button):
    def __init__(self, parent, x, y, normal, pushed, hovering,
            toggle_normal, toggle_pushed, toggle_hovering, toggled=False):

        if toggled:
            super().__init__(parent, x, y, toggle_normal, toggle_pushed,
                             toggle_hovering)
        else:
            super().__init__(parent, x, y, normal, pushed, hovering)

        self.normal = normal
        self.pushed = pushed
        self.hovering = hovering
        self.toggle_normal = toggle_normal
        self.toggle_pushed = toggle_pushed
        self.toggle_hovering = toggle_hovering

        self._toggled_ = toggled

    @property
    def toggled(self):
        return self._toggled_

    @toggled.setter
    def toggled(self, toggled):
        self._toggled_ = toggled

        if toggled:
            self._depressed_img = self.toggle_normal
            self._pressed_img = self.toggle_pushed
            self._hover_img = self.toggle_hovering

        else:
            self._depressed_img = self.normal
            self._pressed_img = self.pushed
            self._hover_img = self.hovering

    def on_press(self):
        self.toggled = not self.toggled
        self.dispatch_event('on_toggle', self.toggled)

ToggleButton.register_event_type('on_toggle')


class TabGroup(Component):
    def __init__(self, parent, x, y, w, h):
        super().__init__(parent, x, y)
        self.size = [w, h]

        self.tabs = [ ]
        self.tabs_width = 0

        pos = self.get_abs_pos()
        self.rect = Rectangle(pos[0], pos[1], w, h, color=(42, 42, 42),
            batch=self.cbatch)

    def set_size(self, w, h):
        if w != self.size[0]:
            self.rect.width = w

        if h != self.size[1]:
            self.rect.height = h

        self.size = [w, h]

    def tab(self, panel, normal, active, hover):
        tab = Tab(self, normal, active, hover)

        if self.tabs:
            panel.hide()

        else:
            panel.show()
            tab._pressed = True
            tab._sprite.image = tab._pressed_img

        self.tabs_width += max(normal.width, active.width, hover.width)
        self.tabs.append((tab, panel))

        return tab


class Tab(Button):
    def __init__(self, tab_group, normal, active, hover):
        x = tab_group.tabs_width
        y = max(normal.height, active.height, hover.height)

        super().__init__(tab_group, x, -y, normal, active, hover)

    def select(self):
        self.on_press()

    def on_press(self):
        for tab, panel in self.parent.tabs:
            if tab is self:
                tab._pressed = True
                tab._sprite.image = tab._pressed_img
                panel.show()

            elif tab._pressed:
                tab._pressed = False
                tab._sprite.image = tab._depressed_img
                panel.hide()

    def on_mouse_release(self, x, y, btns, mods):
        pass




from .__version__ import __version__
from .home import Window as Home
from .node import Client as Node
from .peer import Master as Peer
from .systray import Icon as TrayIcon

import lank.name
import pyglet.app
from PIL import Image, ImageDraw

from threading import Thread
import sys


#icon.notify('this is a test', 'lanku alert')


class Application:
    def __init__(self):
        self.icon = create_image(64, 64, 'green', 'white')
        self.tray_icon = TrayIcon(self)
        self.tray_thread = Thread(target=self.tray_icon.run)

        self.node = None

        self.home = Home(self)

        self.finished = False
        self.exiting = False

    def run(self):
        print(f'lanku v{__version__}')
        try:
            self.tray_thread.start()
            self.show_hide()
            pyglet.app.run()
            self.finished = True

        except KeyboardInterrupt:
            pass

        finally:
            self.quit()
            self.tray_thread.join()

    def show_hide(self):
        if self.home.minimized:
            self.home.set_visible(True)

        else:
            self.home.set_visible(not self.home.visible)

    def quit(self):
        if self.exiting: return
        self.exiting = True
        print('quit')

        self.tray_icon.stop()

        if not self.finished:
            self.home.save()

        self.disconnect()
        pyglet.app.exit()

    def connect(self):
        print('connecting')
        #print(f'pass: {self.home.pnl_config.pwd.value}')

        self.node = Node(self)
        self.node.start()

    def disconnect(self):
        if not self.node: return
        print('disconnecting')
        node = self.node
        self.node = None
        node.stop()
        node.join()
        print('finished')

    def notify(self, signed):
        if signed.name == lank.name.REGISTER:
            self.tray_icon.notify(
                f'{signed.label} has changed keys.',
                f'{signed.label} | lanku')

        elif signed.name == lank.name.PEER:
            self.tray_icon.notify(
                f'{signed.label} has signed on.',
                f'{signed.label} | lanku')

        else:
            self.tray_icon.notify(
                f'{signed.label} has done something.',
                f'{signed.label} | lanku')


def create_image(width, height, color1, color2):
    image = Image.new('RGBA', (width, height), color1)
    dc = ImageDraw.Draw(image)
    dc.rectangle(
        (width // 2, 0, width, height // 2),
        fill=color2)
    dc.rectangle(
        (0, height // 2, width // 2, height),
        fill=color2)

    return image




from . import __version__
from .ui import *
from . import config
from .peers import Panel as Peers

from lank.peer import DEFAULT_PORT


class Window(LWin):
    def __init__(self, app):
        super().__init__(app, 'lanku', 333, 555, save_name='home')
        self.set_minimum_size(300, 250)

        Label(self, 3, 3, f'lanku v{__version__}')

        self.tabs = TabGroup(self, 20, 60, self.xywh[2]-40, self.xywh[3]-100)
        self.pnl_peers = Peers(self.tabs)
        self.pnl_labels = Labels(self.tabs)
        self.pnl_config = Config(self.tabs)

        self.tabs.tab(self.pnl_peers,
            image('tab_peers.png'),
            image('tab_peers-active.png'),
            image('tab_peers-hover.png'))
        self.tabs.tab(self.pnl_labels,
            image('tab_labels.png'),
            image('tab_labels-active.png'),
            image('tab_labels-hover.png'))
        self.tabs.tab(self.pnl_config,
            image('tab_config.png'),
            image('tab_config-active.png'),
            image('tab_config-hover.png')).select()

        self.btn_connect = ToggleButton(self,
            self.xywh[2] / 2 - 58, self.xywh[3] - 35,
            image('btn_connect.png'),
            image('btn_connect-push.png'),
            image('btn_connect-hover.png'),
            image('btn_disconnect.png'),
            image('btn_disconnect-push.png'),
            image('btn_disconnect-hover.png'))
        self.btn_connect.on_toggle = self.on_connect_toggle

    def on_resize(self, w, h):
        super().on_resize(w, h)
        self.tabs.set_size(w-40, h-100)
        self.btn_connect.set_pos(w / 2 - 58, h - 35)

    def on_connect_toggle(self, toggled):
        if toggled:
            self.app.connect()

        else:
            self.app.disconnect()


class Labels(Component):
    def __init__(self, parent):
        super().__init__(parent, 0, 0)

        self.labels = [ ]
        self.interests = config.load_interests()
        #self.refresh_labels(['foo', 'bar'])

    def refresh_labels(self, labels=None):
        clock.schedule_once(self._refresh_labels_, 0, labels)

    def _refresh_labels_(self, dt, labels=None):
        for button, label in self.labels:
            button.remove()
            label.delete()

        self.labels = [ ]

        if labels:
            y = 30
            for label in labels:
                def toggle(label):
                    def toggler(toggled):
                        if toggled:
                            if label not in self.interests:
                                self.interests.append(label)
                        else:
                            if label in self.interests:
                                del self.interests[self.interests.index(label)]
                        config.save_interests(self.interests)
                        if self.win.app.node:
                            self.win.app.node.notify(label, toggled)

                    return toggler

                btn = ToggleButton(self, 3, y,
                    image('btn_bell.png'),
                    image('btn_bell-push.png'),
                    image('btn_bell-hover.png'),
                    image('btn_bell-push.png'),
                    image('btn_bell-hover.png'),
                    image('btn_bell-push.png'),
                    label in self.interests)
                btn.on_toggle = toggle(label)

                self.labels.append((btn, Label(self, 33, y+7, label)))
                y += 30


class Config(Component):
    def __init__(self, parent):
        super().__init__(parent, 0, 0)

        y = 100
        Label(self, 5, y, 'Local Port:')
        self.peer_port = TextEntry(self, 100, y, 75, str(DEFAULT_PORT))
        y -= 40
        Label(self, 5, y, 'Label:')
        self.label = TextEntry(self, 100, y, 150, config.load_connect_label())
        y -= 40
        Label(self, 5, y, 'Password:')
        self.pwd = TextEntry(self, 100, y, 150, mask='*')

